using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Storage;

namespace Frog
{
	class GameController
	{
		//Constants: game parameters
		public const int MaxTeams = 4;
		public const int FrogsPerTeam = 4;
		public const int InitialSpacesPerLane = 9;
		public const int MaxDice = 5;
		public const int MaxZonks = 3;
		public const int MaxMessages = 5;

		private const int BoardLeftMargin = 16;
		private const int BoardRightMargin = 16;
		private const int BoardTopMargin = 32;

		public const int Zonk = 1;

		//Constants: graphics
		private Rectangle BoardPosition;
		private const int DiceWidth = 64;
		private const int MiniDiceWidth = 20;
		private const int ActiveDiceColor = (int)FrogGame.DiceTexture.Green;

		public enum GameState
		{
			NewTurn, PickFrog, TryMove, Zonked, FrogFinish, GameOver
		};

		private Random random;

		//Game state
		private FrogGame game;
		private GameState state;
		private GameBoard board;
		private Team[] teams;
		private List<Team> WinningOrder;
		private Queue<Team> finishOrder;
		private LinkedList<String> messageQueue;

		private int currentTeamIndex;
		private Team currentTeam;
		private int currentLane;
		private int[] dice;
		private int zonks;
		private int spacesAhead;

		public int Zonks
		{
			get { return zonks; }
		}
		public GameState State
		{
			get { return state; }
		}
		public int CurrentTeam
		{
			get { return currentTeamIndex; }
		}

		public GameController(FrogGame game)
		{
			random = new Random();
			this.game = game;

			int rX = BoardLeftMargin + game.safeRect.Left;
			int rY = BoardTopMargin + game.safeRect.Top;

			int rW = game.safeRect.Width - (BoardLeftMargin + BoardRightMargin);
			int rH = 64 * FrogsPerTeam;

			BoardPosition = new Rectangle(rX, rY, rW, rH);

			NewGame();
		}

		public void NewGame()
		{
			//Create a board
			board = new GameBoard(random, FrogsPerTeam, InitialSpacesPerLane);

			//Construct the teams
			teams = new Team[MaxTeams];
			//Establish new list of winners
			WinningOrder = new List<Team>(MaxTeams);
			//Create the queue of messages
			messageQueue = new LinkedList<string>();

			for (int i = 0; i < MaxTeams; i++)
			{
				teams[i] = new Team((Team.TeamID)i, FrogsPerTeam);
				//Give 'em some frogs
				for (int j = 0; j < FrogsPerTeam; j++)
				{
					InitialSpace gs = (InitialSpace)board[j][0];
					Frog f = new Frog(teams[i], gs, j);
					teams[i].AddFrog(f);
					gs.Hold(board, f);
					gs.Accept();
				}
			}
			finishOrder = new Queue<Team>();

			currentTeamIndex = 0;
			dice = new int[MaxDice];
			NewTurn();
		}

		public void PickFrog()
		{
			if (state != GameState.PickFrog)
				return;	//throw new IllegalMoveException();

			spacesAhead += countMoves(currentTeam.ActiveFrog);
			TryMove();
			state = GameState.TryMove;
			AddMessage("Press A (Enter) to roll dice, press B (Space) to accept the move");
		}

		private void NewTurn()
		{
			currentTeam = teams[currentTeamIndex];
			messageQueue.Clear();
			if (state != GameState.FrogFinish)
			{
				AddMessage("Press A (Enter) to roll dice");
			}
			state = GameState.NewTurn;
			currentLane = 0;
			zonks = 0;
			spacesAhead = 0;
			for (int i = 0; i < MaxDice; i++)
				dice[i] = -1;
		}

		private void TryMove()
		{
			Frog f = currentTeam.ActiveFrog;
			foreach (GameSpace s in board[f.ID])
				s.Forget();
			List<GameSpace> lane = board[f.ID];
			lane[f.Space.SpaceNumber].Remove(f);
			if (board.Move(f, f.Space.SpaceNumber + spacesAhead))
			{
				//frog got to the end
				state = GameState.FrogFinish;
				if (currentTeam.IsFinished())
					AddMessage("Your team is finished! Press A (Enter) to continue.");
				else
					AddMessage("Frog finished! Press A (Enter) to begin a new turn");
			}
		}

		public void AcceptMove()
		{
			if (state == GameState.Zonked)
			{
				board.Forget();
			}
			else
			{
				board.Accept();
			}
			currentTeam.ChooseFrog(-1);
			if (state == GameState.FrogFinish)
			{
				if (!currentTeam.IsFinished())
				{
					NewTurn(); //without changing teams
					currentTeam.ChooseFrog(-1);
					RollDice();
				}
				else if (!IsGameOver())
				{
					currentTeam.ChooseFrog(-1);
					WinningOrder.Add(currentTeam);
					do
					{
						currentTeamIndex = (currentTeamIndex + 1) % MaxTeams;
						currentTeam = teams[currentTeamIndex];
					} while (currentTeam.IsFinished());
					state = GameState.NewTurn;
					NewTurn();
				}
				else
				{
					WinningOrder.Add(currentTeam);
					do
					{
						currentTeamIndex = (currentTeamIndex + 1) % MaxTeams;
						currentTeam = teams[currentTeamIndex];
					} while (currentTeam.IsFinished());
					WinningOrder.Add(currentTeam);
					state = GameState.GameOver;
					AddMessage("----Game Over----");
					AddMessage("1st Place: " + WinningOrder[0].ID);
					AddMessage("2nd Place: " + WinningOrder[1].ID);
					AddMessage("3rd Place: " + WinningOrder[2].ID);
					AddMessage("Goat: " + WinningOrder[3].ID);
				}
			}
			else
			{
				if (currentTeam.ActiveFrog == null || !currentTeam.ActiveFrog.Finished)
				{
					do
					{
						currentTeamIndex = (currentTeamIndex + 1) % MaxTeams;
						currentTeam = teams[currentTeamIndex];
					} while (currentTeam.IsFinished());
				}
				NewTurn();
			}
		}

		public void RollDice()
		{
			String diceRolls = new String(' ', 0);
			zonks = 0;
			foreach (int i in dice)
				if (i == Zonk)
					zonks++;
			bool cheat = zonks == MaxZonks - 1 && game.CheckControl(FrogGame.Controls.NoZonks, ButtonState.Pressed);
			for (int i = 0; i < dice.Length; i++)
			{
				if (dice[i] == Zonk)
					continue;

				if (cheat)
				{
					dice[i] = random.Next(board.LaneCount + 1) + 2;
					diceRolls = diceRolls + " " + dice[i];
				}
				else
				{
					dice[i] = random.Next(board.LaneCount + 2) + 1;
					if (dice[i] == Zonk)
					{
						zonks++;
						diceRolls = diceRolls + " Z";
					}
					else if (dice[i] == 6)
					{
						diceRolls = diceRolls + " W";
					}
					else
					{
						diceRolls = diceRolls + " " + dice[i];
					}

					cheat = (cheat || zonks == MaxZonks - 1 && game.CheckControl(FrogGame.Controls.NoZonks, ButtonState.Pressed));
				}
			}
			AddMessage("ROLL: " + diceRolls);

			if (zonks >= MaxZonks)
			{
				state = GameState.Zonked;
				AddMessage("Zonked! Press A (Enter) to continue");
			}
			else if (state == GameState.NewTurn)
			{
				while (!currentTeam.ChooseFrog(currentLane))
					currentLane++;
				state = GameState.PickFrog;
				AddMessage("Use the joystick (Up and Down) to select a frog");
			}
			else
			{
				spacesAhead += countMoves(currentTeam.ActiveFrog);
				state = GameState.TryMove;
				AddMessage("Press A (Enter) to roll dice, press B (Space) to accept the move");
				TryMove();
			}
		}

		public void AddMessage(string stringToAdd)
		{
			while (messageQueue.First != null && messageQueue.Count >= MaxMessages)
			{
				messageQueue.RemoveFirst();
			}
			messageQueue.AddLast(new LinkedListNode<string>(stringToAdd));
		}

		private int countMoves(Frog f)
		{
			int count = 0;
			for (int i = 0; i < dice.Length; i++)
			{
				if (dice[i] == (f.ID + 2) || dice[i] == FrogsPerTeam + 2)
					count++;
			}
			return count;
		}

		public void IncrementLaneChoice(int delta)
		{
			if (state != GameState.PickFrog)
				return; //throw new IllegalMoveException();
			do
			{
				currentLane = (currentLane + delta) % board.LaneCount;
				if (currentLane < 0) currentLane += board.LaneCount;
			} while (!currentTeam.ChooseFrog(currentLane));
		}

		public bool IsGameOver()
		{
			int finishedTeams = 0;
			foreach (Team team in teams)
				if (team.IsFinished())
					finishedTeams++;
			if (finishedTeams >= MaxTeams - 1)
				return true;

			return false;
		}

		public void Draw(FrogGame game, GameTime time)
		{
			SpriteBatch sb = game.SpriteBatch;

			//draw board
			board.Draw(game, BoardPosition, time);

			game.SpriteBatch.Begin(SpriteBlendMode.AlphaBlend);

			//draw dice
			Point diceOffset = new Point(74, 0);
			Rectangle drawRect = new Rectangle((game.safeRect.Width - (DiceWidth * dice.Length)) / 2, BoardPosition.Bottom + 10, DiceWidth, DiceWidth);
			for (int i = 0; i < dice.Length; i++)
			{
				if (dice[i] > 0)
				{
					int color = (int)FrogGame.DiceTexture.White;
					if (dice[i] == currentLane + 2 || dice[i] == 6) color = ActiveDiceColor;
					else if (dice[i] == Zonk) color = (int)FrogGame.DiceTexture.Red;

					sb.Draw(game.dice[dice[i] - 1, color], drawRect, Color.White);
				}
				drawRect.Offset(diceOffset);
			}
			int extraDice = currentTeam.ActiveFrog == null ? 0 : spacesAhead - countMoves(currentTeam.ActiveFrog);
			drawRect = new Rectangle((game.safeRect.Width - (MiniDiceWidth + 5) * extraDice) / 2, drawRect.Bottom + 8, MiniDiceWidth, MiniDiceWidth);
			if (extraDice > 0)
			{
				game.DrawBorder(new Rectangle(drawRect.Left-3,
												drawRect.Top-3,
												(MiniDiceWidth + 5) * extraDice + 3,
												MiniDiceWidth + 6),
								2,
								new Color(0, 0, 0, 196));

				diceOffset = new Point(25, 0);
				for (int i = 0; i < extraDice; i++)
				{
					sb.Draw(game.dice[currentLane + 1, ActiveDiceColor], drawRect, Color.White);
					drawRect.Offset(diceOffset);
				}
			}

			//team status
			game.DrawBorder(new Rectangle(game.safeRect.Left + 10, drawRect.Bottom + 10, 80, 105), 5, new Color(0, 0, 0, 196));
			sb.DrawString(game.font, "Team " + (currentTeamIndex + 1), new Vector2(game.safeRect.Left + 15, drawRect.Bottom + 15), Color.White);
			Frog f = currentTeam.ActiveFrog;
			for (int i = 0; i < 4 && f == null; i++)
				f = currentTeam[i];
			if (f != null)
				f.Draw(game, new Rectangle(game.safeRect.Left + 16, drawRect.Bottom + 40, 70, 70));

			//message queue
			drawRect = new Rectangle(game.safeRect.Left + 100, drawRect.Bottom + 10, game.safeRect.Width - 100 - BoardRightMargin, game.safeRect.Bottom-drawRect.Bottom-20);
			game.DrawBorder(drawRect, 5, new Color(0, 0, 0, 196));
			LinkedListNode<string> currNode = messageQueue.Last;
			for (int i = messageQueue.Count - 1; i >= 0; i--)
			{
				sb.DrawString(game.font, currNode.Value, new Vector2(drawRect.Left + 15, drawRect.Top + i * 20 + 5), currNode == messageQueue.Last ? Color.Red : Color.White);
				currNode = currNode.Previous;
			}

			game.SpriteBatch.End();
		}
	}
}
